/** ** por compatibilidad se omiten tildes **
================================================================================
 TRABAJO PRACTICO 3 - System Programming - ORGANIZACION DE COMPUTADOR II - FCEN
================================================================================

  Definicion de funciones del manejador de memoria
**/

#include "mmu.h"
#include "i386.h"

#include "kassert.h"

static pd_entry_t* kpd = (pd_entry_t*)KERNEL_PAGE_DIR;
static pt_entry_t* kpt = (pt_entry_t*)KERNEL_PAGE_TABLE_0;

static const uint32_t identity_mapping_end = 0x003FFFFF;
static const uint32_t user_memory_pool_end = 0x02FFFFFF;

static paddr_t next_free_kernel_page = 0x100000;
static paddr_t next_free_user_page = 0x400000;   

/**
 * kmemset asigna el valor c a un rango de memoria interpretado
 * como un rango de bytes de largo n que comienza en s
 * @param s es el puntero al comienzo del rango de memoria
 * @param c es el valor a asignar en cada byte de s[0..n-1]
 * @param n es el tamaño en bytes a asignar
 * @return devuelve el puntero al rango modificado (alias de s)
*/
static inline void* kmemset(void* s, int c, size_t n) {
  uint8_t* dst = (uint8_t*)s;
  for (size_t i = 0; i < n; i++) {
    dst[i] = c;
  }
  return dst;
}

/**
 * zero_page limpia el contenido de una página que comienza en addr
 * @param addr es la dirección del comienzo de la página a limpiar
*/
static inline void zero_page(paddr_t addr) {
  kmemset((void*)addr, 0x00, PAGE_SIZE);
}


void mmu_init(void) {}


/**
 * mmu_next_free_kernel_page devuelve la dirección física de la próxima página de kernel disponible. 
 * Las páginas se obtienen en forma incremental, siendo la primera: next_free_kernel_page
 * @return devuelve la dirección de memoria de comienzo de la próxima página libre de kernel
 */
paddr_t mmu_next_free_kernel_page(void) {
  paddr_t temp = next_free_kernel_page;
  next_free_kernel_page += PAGE_SIZE;
  return temp;
}

/**
 * mmu_next_free_user_page devuelve la dirección de la próxima página de usuarix disponible
 * @return devuelve la dirección de memoria de comienzo de la próxima página libre de usuarix
 */
paddr_t mmu_next_free_user_page(void) {
  paddr_t temp = next_free_kernel_page;
  next_free_user_page += PAGE_SIZE;
  return temp;
}

/**
 * mmu_init_kernel_dir inicializa las estructuras de paginación vinculadas al kernel y
 * realiza el identity mapping
 * @return devuelve la dirección de memoria de la página donde se encuentra el directorio
 * de páginas usado por el kernel
 */
paddr_t mmu_init_kernel_dir(void) {
  zero_page((paddr_t)kpd);
  zero_page((paddr_t)kpt);     //vaciamos por si tiene basura

  kpd[0].pt = (uint32_t)kpt >> 12;    //en la primera pos de la kernel page directory apuntamos a la kernel page table (PAGINA 14 DE LA PRACTICA)
  kpd[0].attrs = MMU_P | MMU_W;           //cargamos los atributos que queramos (ver en defines.h)
  for (uint32_t i = 0; i < 1024; i++){    //Tenemos 4MB de id. address y cada page ocupa 4KB (4MB/4KB = 1024)
    kpt[i].page = i;                      //mapeamos las paginas a la funcion identidad
    kpt[i].attrs = MMU_P | MMU_W; 
  }

  return  (paddr_t)kpd;
}

/**
 * mmu_map_page agrega las entradas necesarias a las estructuras de paginación de modo de que
 * la dirección virtual virt se traduzca en la dirección física phy con los atributos definidos en attrs
 * @param cr3 el contenido que se ha de cargar en un registro CR3 al realizar la traducción
 * @param virt la dirección virtual que se ha de traducir en phy
 * @param phy la dirección física que debe ser accedida (dirección de destino)
 * @param attrs los atributos a asignar en la entrada de la tabla de páginas
 */

 //esto mapea(no es solo una traducción), seria como que se apuntan entre si
void mmu_map_page(uint32_t cr3, vaddr_t virt, paddr_t phy, uint32_t attrs) {
  pd_entry_t* pd = (pd_entry_t*)CR3_TO_PAGE_DIR(cr3);
  uint32_t pd_idx = (uint32_t)VIRT_PAGE_DIR(virt);
  uint32_t pt_idx = (uint32_t)VIRT_PAGE_TABLE(virt);

  if((pd[pd_idx].attrs & MMU_P) == 0){
    paddr_t new_pt = mmu_next_free_kernel_page();
    pd[pd_idx].pt = new_pt >> 12;         // paso de usar new_pt como puntero a usarlo como indice de pt
    pd[pd_idx].attrs = MMU_P | attrs;
  }

  pt_entry_t* pte = (pt_entry_t*)MMU_ENTRY_PADDR(pd[pd_idx].pt);
  pte[pt_idx].page = (phy >> 12);
  pte[pt_idx].attrs = MMU_P | attrs;

  tlbflush();           //NO ENTIENDO NADA ESTE CODIGO
}


/**
 * mmu_unmap_page elimina la entrada vinculada a la dirección virt en la tabla de páginas correspondiente
 * @param virt la dirección virtual que se ha de desvincular
 * @return la dirección física de la página desvinculada
 */
paddr_t mmu_unmap_page(uint32_t cr3, vaddr_t virt) {
  pd_entry_t* pd = (pd_entry_t*)CR3_TO_PAGE_DIR(cr3);
  uint32_t pd_idx = (uint32_t)VIRT_PAGE_DIR(virt);
  uint32_t pt_idx = (uint32_t)VIRT_PAGE_TABLE(virt);

  pt_entry_t* pt = (pt_entry_t*)MMU_ENTRY_PADDR(pd[pd_idx].pt);
  paddr_t phy = (paddr_t)MMU_ENTRY_PADDR(pt[pt_idx].page);
  pt[pt_idx].attrs = pt[pt_idx].attrs & (MMU_W | MMU_U);

  return phy;
}

  
#define DST_VIRT_PAGE 0xA00000
#define SRC_VIRT_PAGE 0xB00000

/**
 * copy_page copia el contenido de la página física localizada en la dirección src_addr a la página física ubicada en dst_addr
 * @param dst_addr la dirección a cuya página queremos copiar el contenido
 * @param src_addr la dirección de la página cuyo contenido queremos copiar
 *
 * Esta función mapea ambas páginas a las direcciones SRC_VIRT_PAGE y DST_VIRT_PAGE, respectivamente, realiza
 * la copia y luego desmapea las páginas. Usar la función rcr3 definida en i386.h para obtener el cr3 actual
 */
void copy_page(paddr_t dst_addr, paddr_t src_addr) {
  mmu_map_page(rcr3(), SRC_VIRT_PAGE, src_addr, MMU_P | MMU_W);
  mmu_map_page(rcr3(), DST_VIRT_PAGE, dst_addr, MMU_P | MMU_W);
  uint32_t* src = (uint32_t*)SRC_VIRT_PAGE;
  uint32_t* dst = (uint32_t*)DST_VIRT_PAGE;

  for(uint32_t i = 0; i < 1024; i++){ //el chabon recorre de a 8 bits 4000 veces y nosotros de a 32 1000 veces
    dst[i] = src[i];
  }
  
  mmu_unmap_page(rcr3(),SRC_VIRT_PAGE); 
  mmu_unmap_page(rcr3(),DST_VIRT_PAGE);
}   
 /**
 * mmu_init_task_dir inicializa las estructuras de paginación vinculadas a una tarea cuyo código se encuentra en la dirección phy_start
 * @pararm phy_start es la dirección donde comienzan las dos páginas de código de la tarea asociada a esta llamada
 * @return el contenido que se ha de cargar en un registro CR3 para la tarea asociada a esta llamada
 */

paddr_t mmu_init_task_dir(paddr_t phy_start) {
  paddr_t new_pd = mmu_next_free_kernel_page();
  paddr_t new_pt = mmu_next_free_kernel_page();
  zero_page(new_pd);

  copy_page(new_pd, (paddr_t)kpd);
  copy_page(new_pt, (paddr_t)kpt);

  //Mapeo el codigo
  mmu_map_page(new_pd, (vaddr_t)TASK_CODE_VIRTUAL, phy_start, MMU_P);
  mmu_map_page(new_pd, (vaddr_t)(TASK_CODE_VIRTUAL + PAGE_SIZE), phy_start + PAGE_SIZE, MMU_P);

  //Mapeo el stack
  paddr_t stack_phy = mmu_next_free_user_page();
  mmu_map_page(new_pd, (vaddr_t)(TASK_STACK_BASE - PAGE_SIZE), stack_phy, MMU_P | MMU_W);

  //Mapeo la memoria compartida
  mmu_map_page(new_pd, (vaddr_t)TASK_SHARED_PAGE, stack_phy + PAGE_SIZE, MMU_P);

  tlbflush();
  return new_pd;
}

// COMPLETAR: devuelve true si se atendió el page fault y puede continuar la ejecución 
// y false si no se pudo atender
bool page_fault_handler(vaddr_t virt) {
  print("Atendiendo page fault...       ", 0, 0, C_FG_WHITE | C_BG_BLACK);
  // Chequeemos si el acceso fue dentro del area on-demand
  if (ON_DEMAND_MEM_START_VIRTUAL <= virt  && virt <= ON_DEMAND_MEM_END_VIRTUAL){
    // En caso de que si, mapear la pagina
    mmu_map_page(rcr3(), virt, ON_DEMAND_MEM_START_PHYSICAL, MMU_P | MMU_W | MMU_U);
    return true; 
  } else {
    return false;
  }
}

